/*
 * Copyright 2016 - 2024 Gary Talent (gary@drinkingtea.net). All rights reserved.
 */

#pragma once

#include <ox/std/vector.hpp>
#include <ox/model/def.hpp>

#include <turbine/turbine.hpp>

#include "color.hpp"

namespace nostalgia::core {

struct PaletteColorV1 {
    static constexpr auto TypeName = "net.drinkingtea.nostalgia.core.PaletteColor";
    static constexpr auto TypeVersion = 1;
    uint8_t r{}, g{}, b{}, a{};
    constexpr PaletteColorV1() noexcept = default;
    constexpr PaletteColorV1(Color16 const c) noexcept:
        r{red16(c)},
        g{green16(c)},
        b{blue16(c)},
        a{alpha16(c)} {}
    constexpr operator Color16() const noexcept { return color16(r, g, b, a); }
};

oxModelBegin(PaletteColorV1)
	oxModelField(r)
	oxModelField(g)
	oxModelField(b)
	oxModelField(a)
oxModelEnd()

using PaletteColor = PaletteColorV1;


struct PalettePageV1 {
    static constexpr auto TypeName = "net.drinkingtea.nostalgia.core.Palette.PalettePage";
    static constexpr auto TypeVersion = 1;
    ox::String name;
    ox::Vector<PaletteColorV1> colors;
    constexpr PalettePageV1() noexcept = default;
    constexpr PalettePageV1(ox::StringParam pName, ox::Vector<PaletteColorV1> pColors) noexcept:
        name(std::move(pName)), colors(std::move(pColors)) {}
    constexpr PalettePageV1(ox::StringParam pName, ox::Vector<Color16> const&pColors) noexcept:
        name(std::move(pName)) {
        colors.reserve(pColors.size());
        for (auto const c : pColors) {
            colors.emplace_back(c);
        }
    }
};

oxModelBegin(PalettePageV1)
	oxModelField(name)
	oxModelField(colors)
oxModelEnd()

using PalettePage = PalettePageV1;


struct NostalgiaPalette {
	static constexpr auto TypeName = "net.drinkingtea.nostalgia.core.NostalgiaPalette";
	static constexpr auto TypeVersion = 1;
	ox::Vector<Color16> colors = {};
};

oxModelBegin(NostalgiaPalette)
	oxModelField(colors)
oxModelEnd()


struct PaletteV1 {
	static constexpr auto TypeName = "net.drinkingtea.nostalgia.core.Palette";
	static constexpr auto TypeVersion = 1;
	ox::Vector<Color16> colors;
};

oxModelBegin(PaletteV1)
	oxModelField(colors)
oxModelEnd()


struct PaletteV2 {
	static constexpr auto TypeName = "net.drinkingtea.nostalgia.core.Palette";
	static constexpr auto TypeVersion = 2;
	static constexpr auto Preloadable = true;
	ox::Vector<ox::Vector<Color16>> pages;
};

oxModelBegin(PaletteV2)
	oxModelField(pages)
oxModelEnd()


struct PaletteV3 {
	static constexpr auto TypeName = "net.drinkingtea.nostalgia.core.Palette";
	static constexpr auto TypeVersion = 3;
	static constexpr auto Preloadable = true;
	struct ColorInfo {
		static constexpr auto TypeName = "net.drinkingtea.nostalgia.core.Palette.ColorInfo";
		static constexpr auto TypeVersion = 3;
		ox::String name;
		constexpr ColorInfo() noexcept = default;
		constexpr ColorInfo(ox::StringParam pName) noexcept: name{std::move(pName)} {}
	};
	ox::Vector<ColorInfo> colorInfo;
	ox::Vector<ox::Vector<Color16>> pages;
};

oxModelBegin(PaletteV3::ColorInfo)
	oxModelField(name)
oxModelEnd()

oxModelBegin(PaletteV3)
	oxModelField(colorInfo)
	oxModelField(pages)
oxModelEnd()

[[nodiscard]]
constexpr bool valid(PaletteV3 const&p) noexcept {
	auto const colors = p.colorInfo.size();
	return ox::all_of(p.pages.begin(), p.pages.end(), [colors](auto const&page) {
		return page.size() == colors;
	});
}

constexpr ox::Error repair(PaletteV3 &p) noexcept {
	auto const colors = p.colorInfo.size();
	for (auto &page : p.pages) {
		page.resize(colors);
	}
	return {};
}


struct PaletteV4 {
	static constexpr auto TypeName = "net.drinkingtea.nostalgia.core.Palette";
	static constexpr auto TypeVersion = 4;
	static constexpr auto Preloadable = true;
	ox::Vector<ox::String> colorNames;
	ox::Vector<PalettePageV1> pages;
};

oxModelBegin(PaletteV4)
	oxModelField(colorNames)
	oxModelField(pages)
oxModelEnd()

[[nodiscard]]
constexpr bool valid(PaletteV4 const&p) noexcept {
	auto const colors = p.colorNames.size();
	return ox::all_of(p.pages.begin(), p.pages.end(), [colors](PalettePageV1 const&page) {
		return page.colors.size() == colors;
	});
}

constexpr ox::Error repair(PaletteV4 &p) noexcept {
	auto const colors = p.colorNames.size();
	for (auto &page : p.pages) {
		page.colors.resize(colors);
	}
	return {};
}


using Palette = PaletteV4;


struct CompactPaletteV1 {
	static constexpr auto TypeName = "net.drinkingtea.nostalgia.core.CompactPalette";
	static constexpr auto TypeVersion = 1;
	static constexpr auto Preloadable = true;
	ox::Vector<ox::Vector<Color16>> pages{};
};

oxModelBegin(CompactPaletteV1)
	oxModelField(pages)
oxModelEnd()

[[nodiscard]]
constexpr bool valid(CompactPaletteV1 const&p) noexcept {
	size_t colors{};
	for (auto const&page : p.pages) {
		colors = ox::max(colors, page.size());
	}
	return ox::all_of(p.pages.begin(), p.pages.end(), [colors](ox::Vector<Color16> const&page) {
		return page.size() == colors;
	});
}

constexpr ox::Error repair(CompactPaletteV1 &p) noexcept {
	size_t colors{};
	for (auto const&page : p.pages) {
		colors = ox::max(colors, page.size());
	}
	for (auto &page : p.pages) {
		page.resize(colors);
	}
	return {};
}


using CompactPalette = CompactPaletteV1;


[[nodiscard]]
constexpr Color16 color(Palette const&pal, size_t page, size_t idx) noexcept {
	if (page < pal.pages.size() && idx < pal.pages[page].colors.size()) [[likely]] {
		return pal.pages[page].colors[idx];
	}
	return 0;
}

[[nodiscard]]
constexpr Color16 color(CompactPalette const&pal, size_t page, size_t idx) noexcept {
	if (page < pal.pages.size() && idx < pal.pages[page].size()) [[likely]] {
		return pal.pages[page][idx];
	}
	return 0;
}

[[nodiscard]]
constexpr Color16 color(Palette const&pal, size_t idx) noexcept {
	return color(pal, 0, idx);
}

[[nodiscard]]
constexpr Color16 color(CompactPalette const&pal, size_t idx) noexcept {
	return color(pal, 0, idx);
}

[[nodiscard]]
constexpr auto &colors(Palette &pal, size_t page = 0) noexcept {
	return pal.pages[page].colors;
}

[[nodiscard]]
constexpr auto &colors(CompactPalette &pal, size_t page = 0) noexcept {
	return pal.pages[page];
}

[[nodiscard]]
constexpr auto &colors(Palette const&pal, size_t page = 0) noexcept {
	return pal.pages[page];
}

[[nodiscard]]
constexpr auto &colors(CompactPalette const&pal, size_t page = 0) noexcept {
	return pal.pages[page];
}

[[nodiscard]]
constexpr size_t colorCnt(Palette const&pal, size_t page = 0) noexcept {
	if (page < pal.pages.size()) [[likely]] {
		return pal.pages[page].colors.size();
	}
	return 0;
}

[[nodiscard]]
constexpr size_t colorCnt(CompactPalette const&pal, size_t page = 0) noexcept {
	if (page < pal.pages.size()) [[likely]] {
		return pal.pages[page].size();
	}
	return 0;
}

[[nodiscard]]
constexpr size_t largestPage(Palette const&pal) noexcept {
	size_t out{};
	for (auto const&page : pal.pages) {
		out = ox::max(out, page.colors.size());
	}
	return out;
}

[[nodiscard]]
constexpr size_t largestPage(CompactPalette const&pal) noexcept {
	size_t out{};
	for (auto const&page : pal.pages) {
		out = ox::max(out, page.size());
	}
	return out;
}

}
