/*
 * Copyright 2016 - 2024 Gary Talent (gary@drinkingtea.net). All rights reserved.
 */

#include <ox/fs/fs.hpp>
#include <ox/std/array.hpp>

#include <teagba/addresses.hpp>
#include <teagba/gfx.hpp>
#include <teagba/registers.hpp>

#include <keel/keel.hpp>

#include <nostalgia/core/context.hpp>
#include <nostalgia/core/gfx.hpp>
#include <nostalgia/core/tilesheet.hpp>

#include "context.hpp"

namespace nostalgia::core {

static constexpr auto SpriteCount = 128;

ox::Error initGfx(Context&, InitParams const&) noexcept {
	for (auto bgCtl = &REG_BG0CTL; bgCtl <= &REG_BG3CTL; bgCtl += 2) {
		teagba::bgSetSbb(*bgCtl, 28);
	}
	for (uint16_t i = 0; i < SpriteCount; ++i) {
		auto &sa = teagba::spriteAttr(i);
		sa.idx = i;
	}
	return {};
}

ox::Error loadBgPalette(
		Context&,
		size_t palBank,
		CompactPalette const&palette,
		size_t page) noexcept {
	if (palette.pages.empty()) {
		return {};
	}
	auto const paletteMem = MEM_BG_PALETTE + palBank * 16;
	for (auto i = 0u; i < colorCnt(palette, page); ++i) {
		paletteMem[i] = color(palette, page, i);
	}
	return {};
}

ox::Error loadSpritePalette(
		Context&,
		CompactPalette const&palette,
		size_t page) noexcept {
	if (palette.pages.empty()) {
		return {};
	}
	auto const paletteMem = MEM_SPRITE_PALETTE;
	for (auto i = 0u; i < colorCnt(palette, page); ++i) {
		paletteMem[i] = color(palette, page, i);
	}
	return {};
}

ox::Error loadBgPalette(
		Context &ctx,
		size_t palBank,
		ox::FileAddress const&paletteAddr) noexcept {
	oxRequire(pal, keel::readObj<CompactPalette>(keelCtx(ctx), paletteAddr));
	return loadBgPalette(ctx, palBank, *pal, 0);
}

ox::Error loadSpritePalette(
		Context &ctx,
		ox::FileAddress const&paletteAddr) noexcept {
	oxRequire(pal, keel::readObj<CompactPalette>(keelCtx(ctx), paletteAddr));
	return loadSpritePalette(ctx, *pal, 0);
}

static ox::Error loadTileSheetSet(
		Context &ctx,
		ox::Span<uint16_t> tileMapTargetMem,
		TileSheetSet const&set) noexcept {
	size_t tileWriteIdx = 0;
	size_t const bppMod = set.bpp == 4;
	for (auto const&entry : set.entries) {
		oxRequire(ts, keel::readObj<CompactTileSheet>(keelCtx(ctx), entry.tilesheet));
		if (set.bpp != ts->bpp && ts->bpp == 8) {
			return OxError(1, "cannot load an 8 BPP tilesheet into a 4 BPP CBB");
		}
		for (auto const&s : entry.sections) {
			auto const cnt = (static_cast<size_t>(s.tiles) * PixelsPerTile) >> bppMod;
			for (size_t i = 0; i < cnt; ++i) {
				auto const begin = static_cast<size_t>(s.begin) * (PixelsPerTile >> bppMod);
				auto const srcIdx = begin + i * 2;
				auto const v = static_cast<uint16_t>(
						static_cast<uint16_t>(ts->pixels[srcIdx]) |
				        (static_cast<uint16_t>(ts->pixels[srcIdx + 1]) << 8));
				tileMapTargetMem[tileWriteIdx + i] = v;
			}
			tileWriteIdx += cnt >> bppMod;
		}
	}
	return {};
}

ox::Error loadBgTileSheet(
		Context &ctx,
		unsigned const cbb,
		CompactTileSheet const&ts,
		size_t const dstTileIdx,
		size_t const srcTileIdx,
		size_t const tileCnt) noexcept {
	size_t const bppMod = ts.bpp == 4;
	size_t const bytesPerTile = PixelsPerTile >> bppMod;
	auto const pixCnt = tileCnt * bytesPerTile;
	auto const srcPxIdx = srcTileIdx * bytesPerTile;
	auto const dstPxIdx = (dstTileIdx * bytesPerTile) / 2;
	for (size_t i = 0; i < pixCnt; ++i) {
		auto const srcIdx = srcPxIdx + i * 2;
		auto const p1 = static_cast<uint16_t>(ts.pixels[srcIdx]);
		auto const p2 = static_cast<uint16_t>(ts.pixels[srcIdx + 1]);
		MEM_BG_TILES[cbb][dstPxIdx + i] = static_cast<uint16_t>(p1 | (p2 << 8));
	}
	// update bpp of all bgs with the updated cbb
	auto const bpp = ctx.cbbData[cbb].bpp;
	teagba::iterateBgCtl([bpp, cbb](volatile BgCtl &bgCtl) {
		if (teagba::bgCbb(bgCtl) == cbb) {
			teagba::bgSetBpp(bgCtl, bpp);
		}
	});
	return {};
}

ox::Error loadBgTileSheet(
		Context &ctx,
		unsigned cbb,
		CompactTileSheet const&ts,
		ox::Optional<unsigned> const&paletteBank) noexcept {
	auto const cnt = (ts.pixels.size() * PixelsPerTile) / (1 + (ts.bpp == 4));
	for (size_t i = 0; i < cnt; ++i) {
		auto const srcIdx = i * 2;
		auto const p1 = static_cast<uint16_t>(ts.pixels[srcIdx]);
		auto const p2 = static_cast<uint16_t>(ts.pixels[srcIdx + 1]);
		MEM_BG_TILES[cbb][i] = static_cast<uint16_t>(p1 | (p2 << 8));
	}
	// update bpp of all bgs with the updated cbb
	auto const bpp = ctx.cbbData[cbb].bpp;
	teagba::iterateBgCtl([bpp, cbb](volatile BgCtl &bgCtl) {
		if (teagba::bgCbb(bgCtl) == cbb) {
			teagba::bgSetBpp(bgCtl, bpp);
		}
	});
	if (paletteBank.has_value() && ts.defaultPalette) {
		oxReturnError(loadBgPalette(ctx, *paletteBank, ts.defaultPalette));
	}
	return {};
}

ox::Error loadBgTileSheet(
		Context &ctx,
		unsigned cbb,
		ox::FileAddress const&tilesheetAddr,
		ox::Optional<unsigned> const&paletteBank) noexcept {
	oxRequire(ts, keel::readObj<CompactTileSheet>(keelCtx(ctx), tilesheetAddr));
	return loadBgTileSheet(ctx, cbb, *ts, paletteBank);
}

ox::Error loadBgTileSheet(
		Context &ctx,
		unsigned const cbb,
		TileSheetSet const&set) noexcept {
	auto const bpp = static_cast<unsigned>(set.bpp);
	oxReturnError(loadTileSheetSet(ctx, MEM_BG_TILES[cbb], set));
	// update bpp of all bgs with the updated cbb
	ctx.cbbData[cbb].bpp = bpp;
	teagba::iterateBgCtl([bpp, cbb](volatile BgCtl &bgCtl) {
		if (teagba::bgCbb(bgCtl) == cbb) {
			teagba::bgSetBpp(bgCtl, bpp);
		}
	});
	return {};
}

static void setSpritesBpp(unsigned const bpp) noexcept {
	auto const eightBpp = static_cast<uint16_t >(bpp == 8);
	for (auto i = 0u; i < SpriteCount; ++i) {
		auto &sa = teagba::spriteAttr(i);
		auto &a = sa.attr0;
		a |= static_cast<uint16_t>((a & ~static_cast<uint16_t>(1u << 13)) | (eightBpp << 13));
	}
}

ox::Error loadSpriteTileSheet(
		Context &ctx,
		CompactTileSheet const&ts,
		bool loadDefaultPalette) noexcept {
	for (size_t i = 0; i < ts.pixels.size(); i += 2) {
		uint16_t v = ts.pixels[i];
		v |= static_cast<uint16_t>(ts.pixels[i + 1] << 8);
		MEM_SPRITE_TILES[i] = v;
	}
	if (loadDefaultPalette && ts.defaultPalette) {
		oxReturnError(loadSpritePalette(ctx, ts.defaultPalette));
	}
	setSpritesBpp(static_cast<unsigned>(ts.bpp));
	return {};
}

ox::Error loadSpriteTileSheet(
		Context &ctx,
		ox::FileAddress const&tilesheetAddr,
		bool loadDefaultPalette) noexcept {
	oxRequire(ts, keel::readObj<CompactTileSheet>(keelCtx(ctx), tilesheetAddr));
	return loadSpriteTileSheet(ctx, *ts, loadDefaultPalette);
}

ox::Error loadSpriteTileSheet(
		Context &ctx,
		TileSheetSet const&set) noexcept {
	auto const bpp = static_cast<unsigned>(set.bpp);
	oxReturnError(loadTileSheetSet(ctx, {MEM_SPRITE_TILES, 32 * ox::units::KB}, set));
	setSpritesBpp(bpp);
	return {};
}

void setBgTile(Context &ctx, uint_t bgIdx, int column, int row, BgTile const&tile) noexcept {
	auto const tileIdx = static_cast<std::size_t>(row * tileColumns(ctx) + column);
	// see Tonc 9.3
	MEM_BG_MAP[bgIdx][tileIdx] =
			static_cast<uint16_t>(tile.tileIdx & 0b1'1111'1111) |
			static_cast<uint16_t>(tile.flipX << 0xa) |
			static_cast<uint16_t>(tile.flipY << 0xb) |
			static_cast<uint16_t>(tile.palBank << 0xc);
}

void clearBg(Context &ctx, uint_t bgIdx) noexcept {
	memset(MEM_BG_MAP[bgIdx].data(), 0, static_cast<size_t>(tileRows(ctx) * tileColumns(ctx)));
}

uint8_t bgStatus(Context&) noexcept {
	return (REG_DISPCTL >> 8u) & 0b1111u;
}

void setBgStatus(Context&, uint32_t status) noexcept {
	constexpr auto BgStatus = 8;
	REG_DISPCTL = (REG_DISPCTL & ~0b111100000000u) | status << BgStatus;
}

bool bgStatus(Context&, unsigned bg) noexcept {
	return (REG_DISPCTL >> (8 + bg)) & 1;
}

void setBgStatus(Context&, unsigned bg, bool status) noexcept {
	constexpr auto Bg0Status = 8;
	const auto mask = static_cast<uint32_t>(status) << (Bg0Status + bg);
	REG_DISPCTL = REG_DISPCTL | ((REG_DISPCTL & ~mask) | mask);
}

void setBgBpp(Context&, unsigned bgIdx, unsigned bpp) noexcept {
	auto &bgCtl = regBgCtl(bgIdx);
	teagba::bgSetBpp(bgCtl, bpp);
}

void setBgCbb(Context &ctx, unsigned bgIdx, unsigned cbb) noexcept {
	auto &bgCtl = regBgCtl(bgIdx);
	const auto &cbbData = ctx.cbbData[cbb];
	teagba::bgSetBpp(bgCtl, cbbData.bpp);
	teagba::bgSetCbb(bgCtl, cbb);
}

void setBgPriority(Context&, uint_t bgIdx, uint_t priority) noexcept {
	auto &bgCtl = regBgCtl(bgIdx);
	bgCtl = (bgCtl & 0b1111'1111'1111'1100u) | (priority & 0b11);
}

void hideSprite(Context&, unsigned idx) noexcept {
	//oxAssert(g_spriteUpdates < config::GbaSpriteBufferLen, "Sprite update buffer overflow");
	teagba::addSpriteUpdate({
		.attr0 = uint16_t{0b11 << 8},
		.idx = static_cast<uint16_t>(idx),
	});
}

void showSprite(Context&, unsigned idx) noexcept {
	//oxAssert(g_spriteUpdates < config::GbaSpriteBufferLen, "Sprite update buffer overflow");
	teagba::addSpriteUpdate({
		.attr0 = 0,
		.idx = static_cast<uint16_t>(idx),
	});
}

void setSprite(Context&, uint_t idx, Sprite const&s) noexcept {
	//oxAssert(g_spriteUpdates < config::GbaSpriteBufferLen, "Sprite update buffer overflow");
	uint16_t const eightBpp = s.bpp == 8;
	teagba::addSpriteUpdate({
		.attr0 = static_cast<uint16_t>(
			  (static_cast<uint16_t>(s.y & ox::onMask<uint8_t>(0b111'1111)))
			| (static_cast<uint16_t>(1) << 10) // enable alpha
			| (static_cast<uint16_t>(eightBpp) << 13)
			| (static_cast<uint16_t>(s.spriteShape) << 14)),
		.attr1 = static_cast<uint16_t>(
			  (static_cast<uint16_t>(s.x) & ox::onMask<uint8_t>(8))
			| (static_cast<uint16_t>(s.flipX) << 12)
			| (static_cast<uint16_t>(s.spriteSize) << 14)),
		.attr2 = static_cast<uint16_t>(
			// double tileIdx if 8 bpp
			  (static_cast<uint16_t>((s.tileIdx * (1 + eightBpp)) & ox::onMask<uint16_t>(8)))
			| (static_cast<uint16_t>(s.priority & 0b11) << 10)),
		.idx = static_cast<uint16_t>(idx),
	});
}

uint_t spriteCount(Context&) noexcept {
	return SpriteCount;
}

}
