/*
 * Copyright 2016 - 2024 Gary Talent (gary@drinkingtea.net). All rights reserved.
 */

#include <imgui.h>

#include <keel/media.hpp>

#include <nostalgia/core/gfx.hpp>

#include "paletteeditor.hpp"
#include "paletteeditor-imgui.hpp"

namespace nostalgia::core {

namespace ig = studio::ig;


void PaletteEditorImGui::PageRename::draw(turbine::Context &tctx) noexcept {
	if (!m_show) {
		return;
	}
	if (ig::BeginPopup(tctx, "Rename Page", m_show)) {
		ig::InputText("Name", m_name);
		switch (ig::PopupControlsOkCancel(m_show)) {
			case ig::PopupResponse::OK:
				inputSubmitted.emit(m_name);
				[[fallthrough]];
			case ig::PopupResponse::Cancel:
				close();
			default:
				break;
		}
		ImGui::EndPopup();
	}
}


PaletteEditorImGui::PaletteEditorImGui(studio::StudioContext &sctx, ox::StringParam path):
		Editor(std::move(path)),
		m_sctx(sctx),
		m_tctx(sctx.tctx),
		m_pal(*keel::readObj<Palette>(keelCtx(m_tctx), itemPath()).unwrapThrow()) {
	undoStack()->changeTriggered.connect(this, &PaletteEditorImGui::handleCommand);
	m_pageRename.inputSubmitted.connect(this, &PaletteEditorImGui::renamePage);
}

void PaletteEditorImGui::draw(studio::StudioContext&) noexcept {
	auto const paneSize = ImGui::GetContentRegionAvail();
	{
		ImGui::BeginChild("Pages", {280, paneSize.y}, true);
		drawPagesEditor();
		ImGui::EndChild();
	}
	ImGui::SameLine();
	{
		ImGui::BeginChild("Colors", {-1, paneSize.y}, true);
		drawColorsEditor();
		ImGui::EndChild();
	}
	m_pageRename.draw(m_tctx);
}

ox::Error PaletteEditorImGui::saveItem() noexcept {
	return m_sctx.project->writeObj(itemPath(), m_pal, ox::ClawFormat::Organic);
}

void PaletteEditorImGui::drawColumnLeftAlign(ox::CStringView txt) noexcept {
	ImGui::TableNextColumn();
	ImGui::Text("%s", txt.c_str());
}

void PaletteEditorImGui::drawColumn(ox::CStringView txt) noexcept {
	ImGui::TableNextColumn();
	ImGui::SetCursorPosX(
			ImGui::GetCursorPosX() + ImGui::GetColumnWidth() - ImGui::CalcTextSize(txt.data()).x);
	ImGui::Text("%s", txt.c_str());
}

void PaletteEditorImGui::numShortcuts(size_t &val, size_t const sizeRange) noexcept {
	auto const lastElem = sizeRange - 1;
	if (ImGui::IsKeyPressed(ImGuiKey_0)) {
		val = ox::min<size_t>(9, lastElem);
	} else for (auto i = 9u; i < 10; --i) {
		auto const key = static_cast<ImGuiKey>(ImGuiKey_1 + i);
		if (ImGui::IsKeyPressed(key)) {
			val = ox::min<size_t>(i, lastElem);
			break;
		}
	}
}

void PaletteEditorImGui::colorInput(ox::CStringView label, int &v, bool &inputFocused) noexcept {
	ImGui::InputInt(label.c_str(), &v, 1, 5);
	inputFocused = inputFocused || ImGui::IsItemFocused();
	v = ox::max(v, 0);
}

void PaletteEditorImGui::drawColorsEditor() noexcept {
	constexpr auto tableFlags = ImGuiTableFlags_RowBg;
	auto const colorsSz = ImGui::GetContentRegionAvail();
	auto const colorEditor = m_selectedColorRow < colorCnt(m_pal, m_page);
	auto const colorEditorWidth = 220;
	static constexpr auto toolbarHeight = 40;
	{
		auto constexpr sz = ImVec2{70, 24};
		if (ImGui::Button("Add", sz)) {
			auto const colorSz = colorCnt(m_pal, m_page);
			constexpr Color16 c = 0;
			std::ignore = pushCommand<AddColorCommand>(m_pal, c, colorSz);
		}
		ImGui::SameLine();
		ImGui::BeginDisabled(m_selectedColorRow >= colorCnt(m_pal, m_page));
		{
			if (ImGui::Button("Remove", sz)) {
				std::ignore = pushCommand<RemoveColorCommand>(m_pal, m_selectedColorRow);
				m_selectedColorRow = ox::min(colorCnt(m_pal, m_page) - 1, m_selectedColorRow);
			}
			ImGui::SameLine();
			ImGui::BeginDisabled(m_selectedColorRow <= 0);
			{
				if (ImGui::Button("Move Up", sz)) {
					std::ignore = pushCommand<MoveColorCommand>(
							m_pal, m_page, m_selectedColorRow, m_selectedColorRow - 1);
					--m_selectedColorRow;
				}
			}
			ImGui::EndDisabled();
			ImGui::SameLine();
			ImGui::BeginDisabled(m_selectedColorRow >= colorCnt(m_pal, m_page) - 1);
			{
				if (ImGui::Button("Move Down", sz)) {
					std::ignore = pushCommand<MoveColorCommand>(
							m_pal, m_page, m_selectedColorRow, m_selectedColorRow + 1);
					++m_selectedColorRow;
				}
			}
			ImGui::EndDisabled();
		}
		ImGui::EndDisabled();
	}
	auto const tblWidth = (colorsSz.x - static_cast<float>(colorEditorWidth) - 8.f)
			* static_cast<float>(colorEditor);
	ImGui::BeginTable(
			"Colors",
			6,
			tableFlags,
			{tblWidth, colorsSz.y - (toolbarHeight + 5)});
	{
		ImGui::TableSetupColumn("Idx", ImGuiTableColumnFlags_WidthFixed, 25);
		ImGui::TableSetupColumn("Name", ImGuiTableColumnFlags_WidthFixed, 100);
		ImGui::TableSetupColumn("Red", ImGuiTableColumnFlags_WidthFixed, 40);
		ImGui::TableSetupColumn("Green", ImGuiTableColumnFlags_WidthFixed, 40);
		ImGui::TableSetupColumn("Blue", ImGuiTableColumnFlags_WidthFixed, 40);
		ImGui::TableSetupColumn("Preview", ImGuiTableColumnFlags_NoHide);
		ImGui::TableHeadersRow();
		for (auto i = 0u; auto const&c : m_pal.pages[m_page].colors) {
			ig::IDStackItem const idStackItem(static_cast<int>(i));
			ImGui::TableNextRow();
			drawColumn(i + 1);
			drawColumnLeftAlign(m_pal.colorNames[i]);
			drawColumn(red16(c));
			drawColumn(green16(c));
			drawColumn(blue16(c));
			ImGui::TableNextColumn();
			auto const ic = ImGui::GetColorU32({redf(c), greenf(c), bluef(c), 1});
			ImGui::TableSetBgColor(ImGuiTableBgTarget_CellBg, ic);
			if (ImGui::Selectable(
					"##ColorRow", i == m_selectedColorRow, ImGuiSelectableFlags_SpanAllColumns)) {
				m_selectedColorRow = i;
			}
			++i;
		}
	}
	ImGui::EndTable();
	if (colorEditor) {
		ImGui::SameLine();
		ImGui::BeginChild("ColorEditor", {colorEditorWidth, -1}, true);
		drawColorEditor();
		ImGui::EndChild();
	}
}

void PaletteEditorImGui::drawPagesEditor() noexcept {
	constexpr auto tableFlags = ImGuiTableFlags_RowBg;
	auto const paneSz = ImGui::GetContentRegionAvail();
	constexpr auto toolbarHeight = 40;
	auto const btnSz = ImVec2{paneSz.x / 4 - 5.5f, 24};
	if (ImGui::Button("Add", btnSz)) {
		std::ignore = pushCommand<DuplicatePageCommand>(m_pal, 0u, m_pal.pages.size());
		m_page = m_pal.pages.size() - 1;
	}
	ImGui::SameLine();
	if (ImGui::Button("Remove", btnSz)) {
		std::ignore = pushCommand<RemovePageCommand>(m_pal, m_page);
		m_page = std::min(m_page, m_pal.pages.size() - 1);
	}
	ImGui::SameLine();
	if (ImGui::Button("Clone", btnSz)) {
		std::ignore = pushCommand<DuplicatePageCommand>(m_pal, m_page, m_pal.pages.size());
	}
	ImGui::SameLine();
	if (ImGui::Button("Rename", btnSz)) {
		m_pageRename.show(m_pal.pages[m_page].name);
	}
	ImGui::BeginTable(
			"PageSelect",
			2,
			tableFlags,
			{paneSz.x, paneSz.y - (toolbarHeight + 5)});
	{
		ImGui::TableSetupColumn("Page", ImGuiTableColumnFlags_WidthFixed, 60);
		ImGui::TableSetupColumn("Name", ImGuiTableColumnFlags_WidthFixed, 200);
		ImGui::TableHeadersRow();
		for (auto i = 0u; i < m_pal.pages.size(); ++i) {
			ig::IDStackItem const idStackItem(static_cast<int>(i));
			ImGui::TableNextRow();
			drawColumn(i + 1);
			drawColumnLeftAlign(m_pal.pages[i].name);
			ImGui::SameLine();
			if (ImGui::Selectable("##PageRow", i == m_page, ImGuiSelectableFlags_SpanAllColumns)) {
				m_page = i;
			}
		}
	}
	ImGui::EndTable();
}

void PaletteEditorImGui::drawColorEditor() noexcept {
	auto const c = color(m_pal, m_page, m_selectedColorRow);
	int r = red16(c);
	int g = green16(c);
	int b = blue16(c);
	int const a = alpha16(c);
	auto const newName = ig::InputText<50>(
			"Name", m_pal.colorNames[m_selectedColorRow]);
	bool inputFocused = ImGui::IsItemFocused();
	ImGui::Separator();
	colorInput("Red", r, inputFocused);
	colorInput("Green", g, inputFocused);
	colorInput("Blue", b, inputFocused);
	if (ig::PushButton("Apply to all pages", {-1, ig::BtnSz.y})) {
		std::ignore = pushCommand<ApplyColorAllPagesCommand>(
				m_pal, m_page, m_selectedColorRow);
	}
	if (!inputFocused && !m_pageRename.isOpen()) {
		if (!ImGui::IsKeyDown(ImGuiKey_ModAlt)) {
			numShortcuts(m_selectedColorRow, largestPage(m_pal));
		} else {
			numShortcuts(m_page, m_pal.pages.size());
		}
	}
	auto const newColor = color16(r, g, b, a);
	if (c != newColor) {
		std::ignore = pushCommand<UpdateColorCommand>(m_pal, m_page, m_selectedColorRow, newColor);
	}
	if (newName) {
		std::ignore = pushCommand<UpdateColorInfoCommand>(
				m_pal, m_selectedColorRow, ox::String{newName});
	}
}

ox::Error PaletteEditorImGui::renamePage(ox::StringView name) noexcept {
	return pushCommand<RenamePageCommand>(m_pal, m_page, name);
}

ox::Error PaletteEditorImGui::handleCommand(studio::UndoCommand const*cmd) noexcept {
	if (dynamic_cast<RemovePageCommand const*>(cmd)) {
		m_page = ox::min(m_page, m_pal.pages.size() - 1);
	} else if (auto const dupPageCmd = dynamic_cast<DuplicatePageCommand const*>(cmd)) {
		m_page = ox::clamp<size_t>(dupPageCmd->insertIdx(), 0, m_pal.pages.size() - 1);
	}
	return {};
}

}
