/*
 * Copyright 2016 - 2024 Gary Talent (gary@drinkingtea.net). All rights reserved.
 */

#include "paletteeditor.hpp"

namespace nostalgia::core {

ApplyColorAllPagesCommand::ApplyColorAllPagesCommand(Palette &pal, size_t const page, size_t const idx):
	m_pal(pal),
	m_page(page),
	m_idx(idx),
	m_origColors([this] {
		ox::Vector<Color16> colors;
		colors.reserve(m_pal.pages.size());
		for (auto const&p : m_pal.pages) {
			colors.emplace_back(p.colors[m_idx]);
		}
		return colors;
	}()) {
	auto const c = color(m_pal, m_page, m_idx);
	if (ox::all_of(m_pal.pages.begin(), m_pal.pages.end(), [this, c](PalettePage const&page) {
			return page.colors[m_idx] == c;
		})) {
		throw studio::NoChangesException();
	}
}

int ApplyColorAllPagesCommand::commandId() const noexcept {
	return static_cast<int>(PaletteEditorCommandId::ApplyColorAllPages);
}

ox::Error ApplyColorAllPagesCommand::redo() noexcept {
	auto const c = color(m_pal, m_page, m_idx);
	for (auto &page : m_pal.pages) {
		page.colors[m_idx] = c;
	}
	return {};
}

ox::Error ApplyColorAllPagesCommand::undo() noexcept {
	for (size_t p = 0u; auto &page : m_pal.pages) {
		page.colors[m_idx] = m_origColors[p];
		++p;
	}
	return {};
}


RenamePageCommand::RenamePageCommand(Palette &pal, size_t const page, ox::StringParam name) noexcept:
		m_pal(pal),
		m_page(page),
		m_name{std::move(name)} {}

int RenamePageCommand::commandId() const noexcept {
	return static_cast<int>(PaletteEditorCommandId::RenamePage);
}

ox::Error RenamePageCommand::redo() noexcept {
	std::swap(m_pal.pages[m_page].name, m_name);
	return {};
}

ox::Error RenamePageCommand::undo() noexcept {
	std::swap(m_pal.pages[m_page].name, m_name);
	return {};
}


DuplicatePageCommand::DuplicatePageCommand(Palette &pal, size_t srcIdx, size_t dstIdx) noexcept:
		m_pal(pal),
		m_dstIdx(dstIdx) {
	auto const&src = m_pal.pages[srcIdx];
	m_page.reserve(src.colors.size());
	for (auto const&s : src.colors) {
		m_page.emplace_back(s);
	}
}

int DuplicatePageCommand::commandId() const noexcept {
	return static_cast<int>(PaletteEditorCommandId::DuplicatePage);
}

ox::Error DuplicatePageCommand::redo() noexcept {
	m_pal.pages.emplace(m_dstIdx, "", std::move(m_page));
	return {};
}

ox::Error DuplicatePageCommand::undo() noexcept {
	m_page = std::move(m_pal.pages[m_dstIdx].colors);
	return m_pal.pages.erase(m_dstIdx).error;
}

size_t DuplicatePageCommand::insertIdx() const noexcept {
	return m_dstIdx;
}


RemovePageCommand::RemovePageCommand(Palette &pal, size_t idx) noexcept:
	m_pal(pal),
	m_idx(idx) {}

int RemovePageCommand::commandId() const noexcept {
	return static_cast<int>(PaletteEditorCommandId::RemovePage);
}

ox::Error RemovePageCommand::redo() noexcept {
	m_page = std::move(m_pal.pages[m_idx]);
	return m_pal.pages.erase(m_idx).error;
}

ox::Error RemovePageCommand::undo() noexcept {
	m_pal.pages.emplace(m_idx, std::move(m_page));
	return {};
}


AddColorCommand::AddColorCommand(Palette &pal, Color16 const color, size_t const idx) noexcept:
	m_pal(pal),
	m_color(color),
	m_idx(idx) {}

int AddColorCommand::commandId() const noexcept {
	return static_cast<int>(PaletteEditorCommandId::AddColor);
}

ox::Error AddColorCommand::redo() noexcept {
	m_pal.colorNames.emplace(m_idx, ox::sfmt("Color {}", m_pal.colorNames.size() + 1));
	for (auto &page : m_pal.pages) {
		page.colors.emplace(m_idx, m_color);
	}
	return {};
}

ox::Error AddColorCommand::undo() noexcept {
	oxReturnError(m_pal.colorNames.erase(m_idx));
	for (auto &page : m_pal.pages) {
		oxReturnError(page.colors.erase(m_idx));
	}
	return {};
}


RemoveColorCommand::RemoveColorCommand(Palette &pal, size_t const idx) noexcept:
	m_pal(pal),
	m_idx(idx),
	m_colors([this] {
		ox::Vector<Color16> colors;
		colors.reserve(m_pal.pages.size());
		for (auto const&p : m_pal.pages) {
			colors.emplace_back(p.colors[m_idx]);
		}
		return colors;
	}()) {}

int RemoveColorCommand::commandId() const noexcept {
	return static_cast<int>(PaletteEditorCommandId::RemoveColor);
}

ox::Error RemoveColorCommand::redo() noexcept {
	m_colorInfo = std::move(m_pal.colorNames[m_idx]);
	oxReturnError(m_pal.colorNames.erase(m_idx));
	for (auto &page : m_pal.pages) {
		oxReturnError(page.colors.erase(m_idx));
	}
	return {};
}

ox::Error RemoveColorCommand::undo() noexcept {
	m_pal.colorNames.emplace(m_idx, std::move(m_colorInfo));
	for (size_t p = 0; auto &page : m_pal.pages) {
		page.colors.emplace(m_idx, m_colors[p]);
		++p;
	}
	return {};
}


UpdateColorInfoCommand::UpdateColorInfoCommand(
		Palette &pal,
		size_t idx,
		ox::StringParam newColorInfo):
		m_pal(pal),
		m_idx(idx),
		m_altColorInfo(std::move(newColorInfo)) {
	if (m_pal.colorNames[m_idx] == m_altColorInfo) {
		throw studio::NoChangesException();
	}
}

bool UpdateColorInfoCommand::mergeWith(UndoCommand &cmd) noexcept {
	if (cmd.commandId() != static_cast<int>(PaletteEditorCommandId::UpdateColorInfo)) {
		return false;
	}
	auto ucCmd = dynamic_cast<UpdateColorInfoCommand const*>(&cmd);
	if (m_idx != ucCmd->m_idx) {
		return false;
	}
	m_pal.colorNames[m_idx] = std::move(ucCmd->m_pal.colorNames[m_idx]);
	setObsolete(m_altColorInfo == m_pal.colorNames[m_idx]);
	return true;
}

int UpdateColorInfoCommand::commandId() const noexcept {
	return static_cast<int>(PaletteEditorCommandId::UpdateColorInfo);
}

ox::Error UpdateColorInfoCommand::redo() noexcept {
	swap();
	return {};
}

ox::Error UpdateColorInfoCommand::undo() noexcept {
	swap();
	return {};
}

void UpdateColorInfoCommand::swap() noexcept {
	std::swap(m_pal.colorNames[m_idx], m_altColorInfo);
}


UpdateColorCommand::UpdateColorCommand(
	Palette &pal,
	size_t page,
	size_t idx,
	Color16 newColor):
		m_pal(pal),
		m_page(page),
		m_idx(idx),
		m_altColor(newColor) {
	if (color(m_pal, m_page, m_idx) == newColor) {
		throw studio::NoChangesException();
	}
}

bool UpdateColorCommand::mergeWith(UndoCommand &cmd) noexcept {
	if (cmd.commandId() != static_cast<int>(PaletteEditorCommandId::UpdateColor)) {
		return false;
	}
	auto ucCmd = dynamic_cast<UpdateColorCommand const*>(&cmd);
	if (m_idx != ucCmd->m_idx) {
		return false;
	}
	return true;
}

int UpdateColorCommand::commandId() const noexcept {
	return static_cast<int>(PaletteEditorCommandId::UpdateColor);
}

ox::Error UpdateColorCommand::redo() noexcept {
	swap();
	return {};
}

ox::Error UpdateColorCommand::undo() noexcept {
	swap();
	return {};
}

void UpdateColorCommand::swap() noexcept {
	auto &dst = colors(m_pal, m_page)[m_idx];
	std::swap(dst, m_altColor);
}


MoveColorCommand::MoveColorCommand(
		Palette &pal, size_t page, size_t srcIdx, size_t dstIdx) noexcept:
	m_pal(pal),
	m_page(page),
	m_srcIdx(srcIdx),
	m_dstIdx(dstIdx) {}

int MoveColorCommand::commandId() const noexcept {
	return static_cast<int>(PaletteEditorCommandId::MoveColor);
}

ox::Error MoveColorCommand::redo() noexcept {
	moveColor(m_srcIdx, m_dstIdx);
	return {};
}

ox::Error MoveColorCommand::undo() noexcept {
	moveColor(m_dstIdx, m_srcIdx);
	return {};
}

void MoveColorCommand::moveColor(size_t srcIdx, size_t dstIdx) noexcept {
	auto const c = color(m_pal, m_page, srcIdx);
	std::ignore = colors(m_pal, m_page).erase(srcIdx);
	colors(m_pal, m_page).emplace(dstIdx, c);
}

}
