/*
 * Copyright 2016 - 2024 Gary Talent (gary@drinkingtea.net). All rights reserved.
 */

#pragma once

#include <studio/studio.hpp>

#include <nostalgia/core/color.hpp>
#include <nostalgia/core/palette.hpp>

namespace nostalgia::core {

enum class PaletteEditorCommandId {
	ApplyColorAllPages,
	RenamePage,
	DuplicatePage,
	RemovePage,
	AddColor,
	RemoveColor,
	UpdateColorInfo,
	UpdateColor,
	MoveColor,
};


class ApplyColorAllPagesCommand: public studio::UndoCommand {
	private:
		Palette &m_pal;
		size_t const m_page{};
		size_t const m_idx{};
		ox::Vector<Color16> const m_origColors;

	public:
		ApplyColorAllPagesCommand(Palette &pal, size_t page, size_t idx);

		~ApplyColorAllPagesCommand() noexcept override = default;

		[[nodiscard]]
		int commandId() const noexcept final;

		ox::Error redo() noexcept final;

		ox::Error undo() noexcept final;
};

class RenamePageCommand: public studio::UndoCommand {
	private:
		Palette &m_pal;
		size_t m_page = 0;
		ox::String m_name;

	public:
		RenamePageCommand(Palette &pal, size_t page, ox::StringParam name) noexcept;

		~RenamePageCommand() noexcept override = default;

		[[nodiscard]]
		int commandId() const noexcept final;

		ox::Error redo() noexcept final;

		ox::Error undo() noexcept final;

};

class DuplicatePageCommand: public studio::UndoCommand {
	private:
		Palette &m_pal;
		size_t m_dstIdx = 0;
		ox::Vector<PaletteColor> m_page;

	public:
		DuplicatePageCommand(Palette &pal, size_t srcIdx, size_t dstIdx) noexcept;

		~DuplicatePageCommand() noexcept override = default;

		[[nodiscard]]
		int commandId() const noexcept final;

		ox::Error redo() noexcept final;

		ox::Error undo() noexcept final;

		[[nodiscard]]
		size_t insertIdx() const noexcept;

};

class RemovePageCommand: public studio::UndoCommand {
	private:
		Palette &m_pal;
		size_t m_idx = 0;
		PalettePage m_page;

	public:
		RemovePageCommand(Palette &pal, size_t idx) noexcept;

		~RemovePageCommand() noexcept override = default;

		[[nodiscard]]
		int commandId() const noexcept final;

		ox::Error redo() noexcept final;

		ox::Error undo() noexcept final;

};

class AddColorCommand: public studio::UndoCommand {
	private:
		Palette &m_pal;
		Color16 m_color = 0;
		size_t const m_idx = 0;

	public:
		AddColorCommand(Palette &pal, Color16 color, size_t idx) noexcept;

		~AddColorCommand() noexcept override = default;

		[[nodiscard]]
		int commandId() const noexcept override;

		ox::Error redo() noexcept override;

		ox::Error undo() noexcept override;

};

class RemoveColorCommand: public studio::UndoCommand {
	private:
		Palette &m_pal;
		size_t const m_idx = 0;
		ox::String m_colorInfo;
		ox::Vector<Color16> const m_colors;

	public:
		RemoveColorCommand(Palette &pal, size_t idx) noexcept;

		~RemoveColorCommand() noexcept override = default;

		[[nodiscard]]
		int commandId() const noexcept override;

		ox::Error redo() noexcept override;

		ox::Error undo() noexcept override;

};

class UpdateColorInfoCommand: public studio::UndoCommand {
	private:
		Palette &m_pal;
		size_t const m_idx{};
		ox::String m_altColorInfo;

	public:
		UpdateColorInfoCommand(
				Palette &pal,
				size_t idx,
				ox::StringParam newColorInfo);

		~UpdateColorInfoCommand() noexcept override = default;

		[[nodiscard]]
		bool mergeWith(UndoCommand &cmd) noexcept final;

		[[nodiscard]]
		int commandId() const noexcept final;

		ox::Error redo() noexcept final;

		ox::Error undo() noexcept final;

	private:
		void swap() noexcept;

};

class UpdateColorCommand: public studio::UndoCommand {
	private:
		Palette &m_pal;
		size_t const m_page = 0;
		size_t const m_idx{};
		PaletteColor m_altColor{};

	public:
		UpdateColorCommand(
				Palette &pal,
				size_t page,
				size_t idx,
				Color16 newColor);

		~UpdateColorCommand() noexcept override = default;

		[[nodiscard]]
		bool mergeWith(UndoCommand &cmd) noexcept final;

		[[nodiscard]]
		int commandId() const noexcept final;

		ox::Error redo() noexcept final;

		ox::Error undo() noexcept final;

	private:
		void swap() noexcept;

};

class MoveColorCommand: public studio::UndoCommand {
	private:
		Palette &m_pal;
		size_t const m_page = 0;
		std::size_t const m_srcIdx = 0;
		std::size_t const m_dstIdx = 0;

	public:
		MoveColorCommand(Palette &pal, size_t page, size_t srcIdx, size_t dstIdx) noexcept;

		~MoveColorCommand() noexcept override = default;

		[[nodiscard]]
		int commandId() const noexcept override;

		ox::Error redo() noexcept override;

		ox::Error undo() noexcept override;

	private:
		void moveColor(size_t srcIdx, size_t dstIdx) noexcept;
};

}
