/*
 * Copyright 2016 - 2024 Gary Talent (gary@drinkingtea.net). All rights reserved.
 */

#pragma once

#include <ox/std/defines.hpp>

#include <ox/claw/claw.hpp>
#include <ox/fs/fs.hpp>
#include <ox/model/metadata.hpp>


#include "asset.hpp"
#include "context.hpp"
#include "typeconv.hpp"

namespace keel {

// Pointer to preloaded data that can be stored in FS in place of the actual
// data.
struct PreloadPtr {
	static constexpr auto TypeName = "net.drinkingtea.keel.PreloadPtr";
	static constexpr auto TypeVersion = 2;
	uint64_t preloadAddr = 0;
};

oxModelBegin(PreloadPtr)
	oxModelField(preloadAddr)
oxModelEnd()

ox::Result<std::size_t> getPreloadAddr(keel::Context &ctx, ox::FileAddress const&file) noexcept;
ox::Result<std::size_t> getPreloadAddr(keel::Context &ctx, ox::StringViewCR file) noexcept;


void createUuidMapping(Context &ctx, ox::StringView filePath, ox::UUID const&uuid) noexcept;

ox::Error buildUuidMap(Context &ctx) noexcept;

ox::Result<ox::UUID> pathToUuid(Context &ctx, ox::StringViewCR path) noexcept;

ox::Result<ox::UUID> getUuid(Context &ctx, ox::FileAddress const&fileAddr) noexcept;

ox::Result<ox::UUID> getUuid(Context &ctx, ox::StringView path) noexcept;

ox::Result<ox::CStringView> getPath(Context &ctx, ox::FileAddress const&fileAddr) noexcept;

ox::Result<ox::CStringView> getPath(Context &ctx, ox::CStringView fileId) noexcept;

constexpr ox::Result<ox::UUID> uuidUrlToUuid(ox::StringView uuidUrl) noexcept {
	return ox::UUID::fromString(substr(uuidUrl, 7));
}

ox::Result<ox::CStringView> uuidUrlToPath(Context &ctx, ox::StringView uuid) noexcept;

ox::Result<ox::CStringView> uuidToPath(Context &ctx, ox::StringViewCR uuid) noexcept;

ox::Result<ox::CStringView> uuidToPath(Context &ctx, ox::UUID const&uuid) noexcept;

#ifndef OX_BARE_METAL

namespace detail {
template<typename T>
constexpr auto makeLoader(Context &ctx) {
	return [&ctx](ox::StringView assetId) -> ox::Result<T> {
		ox::StringView path;
		oxRequire(p, ctx.uuidToPath.at(assetId));
		path = *p;
		oxRequire(buff, ctx.rom->read(path));
		auto [obj, err] = readAsset<T>(buff);
		if (err) {
			if (err != ox::Error_ClawTypeVersionMismatch && err != ox::Error_ClawTypeMismatch) {
				return err;
			}
			oxReturnError(convert<T>(ctx, buff, &obj));
		}
		return std::move(obj);
	};
};
}

template<typename T>
ox::Result<keel::AssetRef<T>> readObjFile(
		keel::Context &ctx,
		ox::StringView assetId,
		bool forceLoad) noexcept {
	ox::UUIDStr uuidStr;
	if (beginsWith(assetId, "uuid://")) {
		assetId = substr(assetId, 7);
		oxRequire(p, keel::uuidToPath(ctx, assetId));
	} else {
		auto const [uuid, uuidErr] = getUuid(ctx, assetId);
		if (!uuidErr) {
			uuidStr = uuid.toString();
			assetId = uuidStr;
		}
	}
	if (forceLoad) {
		ctx.assetManager.initTypeManager<T>(detail::makeLoader<T>, ctx);
		oxRequire(cached, ctx.assetManager.loadAsset<T>(assetId));
		return cached;
	} else {
		auto [cached, err] = ctx.assetManager.getAsset<T>(assetId);
		if (err) {
			ctx.assetManager.initTypeManager<T>(detail::makeLoader<T>, ctx);
			oxReturnError(ctx.assetManager.loadAsset<T>(assetId).moveTo(cached));
		}
		return cached;
	}
}

#else

template<typename T>
ox::Result<keel::AssetRef<T>> readObjNoCache(
		keel::Context &ctx,
		ox::StringViewCR assetId) noexcept {
	if constexpr(ox::preloadable<T>::value) {
		oxRequire(addr, getPreloadAddr(ctx, assetId));
		return keel::AssetRef<T>(std::bit_cast<T const*>(uintptr_t{addr}));
	} else {
		return OxError(1);
	}
}

#endif

ox::Error reloadAsset(keel::Context &ctx, ox::StringView assetId) noexcept;

template<typename T>
ox::Result<keel::AssetRef<T>> readObj(
		keel::Context &ctx,
		ox::StringViewCR assetId,
		[[maybe_unused]] bool forceLoad = false) noexcept {
#ifndef OX_BARE_METAL
	return readObjFile<T>(ctx, assetId, forceLoad);
#else
	return readObjNoCache<T>(ctx, assetId);
#endif
}

template<typename T>
ox::Result<keel::AssetRef<T>> readObj(
		keel::Context &ctx,
		ox::FileAddress const&file,
		[[maybe_unused]] bool forceLoad = false) noexcept {
#ifndef OX_BARE_METAL
	oxRequire(assetId, file.getPath());
	return readObj<T>(ctx, ox::StringView(assetId), forceLoad);
#else
	if constexpr(ox::preloadable<T>::value) {
		oxRequire(addr, getPreloadAddr(ctx, file));
		return keel::AssetRef<T>(std::bit_cast<T const*>(uintptr_t{addr}));
	} else {
		return OxError(1);
	}
#endif
}

template<typename T>
ox::Error writeObj(
		keel::Context &ctx,
		ox::FileAddress const&file,
		T const&obj,
		ox::ClawFormat fmt = ox::ClawFormat::Metal) noexcept {
	oxRequire(objBuff, ox::writeClaw(obj, fmt));
	return ctx.rom->write(file, objBuff.data(), objBuff.size());
}

ox::Error setRomFs(Context &ctx, ox::UPtr<ox::FileSystem> &&fs) noexcept;

ox::Result<ox::UniquePtr<ox::FileSystem>> loadRomFs(ox::StringViewCR path) noexcept;

ox::Result<char*> loadRom(ox::StringViewCR path = "") noexcept;

void unloadRom(char*) noexcept;

}
