/*
 * Copyright 2016 - 2024 Gary Talent (gary@drinkingtea.net). All rights reserved.
 */

#pragma once

#include <ox/std/def.hpp>
#include <ox/std/error.hpp>
#include <ox/std/string.hpp>
#include <ox/claw/read.hpp>
#include <ox/claw/write.hpp>

#include "asset.hpp"
#include "context.hpp"

namespace keel {

class Wrap {
	public:
		virtual ~Wrap() = default;
};

template<typename T>
class WrapInline: public Wrap {
	private:
		T m_obj;

	public:
		constexpr WrapInline() = default;

		template<typename... Args>
		constexpr explicit WrapInline(Args &&...args): m_obj(ox::forward<Args>(args)...) {
		}

		[[nodiscard]]
		constexpr T &obj() noexcept {
			return m_obj;
		}

};

template<typename T, typename... Args>
constexpr ox::UPtr<Wrap> makeWrap(Args &&...args) noexcept {
	return ox::make_unique<WrapInline<T>>(ox::forward<Args>(args)...);
}

template<typename T>
constexpr T &wrapCast(Wrap &ptr) noexcept {
	return static_cast<WrapInline<T>&>(ptr).obj();
}

class BaseConverter {
	public:
		constexpr virtual ~BaseConverter() noexcept = default;

		[[nodiscard]]
		constexpr virtual ox::StringView srcTypeName() const noexcept = 0;

		[[nodiscard]]
		constexpr virtual int srcTypeVersion() const noexcept = 0;

		[[nodiscard]]
		constexpr virtual bool srcMatches(ox::StringViewCR pSrcTypeName, int pSrcTypeVersion) const noexcept = 0;

		[[nodiscard]]
		constexpr virtual bool dstMatches(ox::StringViewCR dstTypeName, int dstTypeVersion) const noexcept = 0;

		virtual ox::Result<ox::UPtr<Wrap>> convertPtrToPtr(keel::Context &ctx, Wrap &src) const noexcept = 0;

		virtual ox::Result<ox::UPtr<Wrap>> convertBuffToPtr(
				keel::Context &ctx, ox::BufferView const&srcBuff) const noexcept = 0;

		[[nodiscard]]
		constexpr bool matches(
				ox::StringViewCR srcTypeName, int srcTypeVersion,
				ox::StringViewCR dstTypeName, int dstTypeVersion) const noexcept {
			return srcMatches(srcTypeName, srcTypeVersion)
				 && dstMatches(dstTypeName, dstTypeVersion);
		}

};

template<typename SrcType, typename DstType>
class Converter: public BaseConverter {
	public:
		[[nodiscard]]
		constexpr ox::StringView srcTypeName() const noexcept final {
			return ox::ModelTypeName_v<SrcType>;
		}

		[[nodiscard]]
		constexpr int srcTypeVersion() const noexcept final {
			return ox::ModelTypeVersion_v<SrcType>;
		}

		[[nodiscard]]
		constexpr bool srcMatches(ox::StringViewCR pSrcTypeName, int pSrcTypeVersion) const noexcept final {
			constexpr auto SrcTypeName = ox::requireModelTypeName<SrcType>();
			constexpr auto SrcTypeVersion = ox::requireModelTypeVersion<SrcType>();
			return pSrcTypeName == SrcTypeName
			    && pSrcTypeVersion == SrcTypeVersion;
		}

		[[nodiscard]]
		constexpr bool dstMatches(ox::StringViewCR dstTypeName, int dstTypeVersion) const noexcept final {
			constexpr auto DstTypeName = ox::StringView{ox::requireModelTypeName<DstType>()};
			constexpr auto DstTypeVersion = ox::requireModelTypeVersion<DstType>();
			return dstTypeName == DstTypeName
			    && dstTypeVersion == DstTypeVersion;
		}

		ox::Result<ox::UPtr<Wrap>> convertPtrToPtr(
				keel::Context &ctx, Wrap &src) const noexcept final {
			auto dst = makeWrap<DstType>();
			oxReturnError(convert(ctx, wrapCast<SrcType>(src), wrapCast<DstType>(*dst)));
			return {std::move(dst)};
		}

		ox::Result<ox::UPtr<Wrap>> convertBuffToPtr(
				keel::Context &ctx, ox::BufferView const&srcBuff) const noexcept final {
			oxRequireM(src, readAsset<SrcType>(srcBuff));
			oxReturnError(ensureValid(src));
			auto dst = makeWrap<DstType>();
			oxReturnError(convert(ctx, src, wrapCast<DstType>(*dst)));
			return {std::move(dst)};
		}

	protected:
		virtual ox::Error convert(keel::Context &ctx, SrcType&, DstType&) const noexcept = 0;

};

ox::Result<ox::UPtr<Wrap>> convert(
		keel::Context &ctx,
		ox::BufferView const&srcBuffer,
		ox::StringViewCR dstTypeName,
		int dstTypeVersion) noexcept;

template<typename DstType>
ox::Result<DstType> convert(keel::Context &ctx, ox::BufferView const&srcBuffer) noexcept {
	static constexpr auto DstTypeName = ox::requireModelTypeName<DstType>();
	static constexpr auto DstTypeVersion = ox::requireModelTypeVersion<DstType>();
	oxRequire(out, convert(ctx, srcBuffer, DstTypeName, DstTypeVersion));
	return std::move(wrapCast<DstType>(out));
}

template<typename DstType>
ox::Error convert(keel::Context &ctx, ox::BufferView const&buff, DstType *outObj) noexcept {
	static constexpr auto DstTypeName = ox::requireModelTypeName<DstType>();
	static constexpr auto DstTypeVersion = ox::requireModelTypeVersion<DstType>();
	oxRequire(outPtr, convert(ctx, buff, DstTypeName, DstTypeVersion));
	*outObj = std::move(wrapCast<DstType>(*outPtr));
	return {};
}

template<typename DstType>
ox::Result<ox::Buffer> convertBuffToBuff(
		keel::Context &ctx, ox::BufferView const&srcBuffer, ox::ClawFormat fmt) noexcept {
	static constexpr auto DstTypeName = ox::requireModelTypeName<DstType>();
	static constexpr auto DstTypeVersion = ox::requireModelTypeVersion<DstType>();
	oxRequire(out, convert(ctx, srcBuffer, DstTypeName, DstTypeVersion));
	return ox::writeClaw<DstType>(wrapCast<DstType>(*out), fmt);
}

template<typename From, typename To, ox::ClawFormat fmt = ox::ClawFormat::Metal>
ox::Result<bool> transformRule(keel::Context &ctx, ox::Buffer &buff, ox::StringView typeId) noexcept {
	if (typeId == ox::ModelTypeId_v<From>) {
		oxReturnError(keel::convertBuffToBuff<To>(ctx, buff, fmt).moveTo(buff));
		return true;
	}
	return false;
};


}
